/*
 *      casCoreClientIL.h,v 1.3 1999/09/02 21:50:08 jhill Exp
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 *
 * History
 * casCoreClientIL.h,v
 * Revision 1.3  1999/09/02 21:50:08  jhill
 * o changed UDP to non-blocking IO
 * o cleaned up (consolodated) UDP interface class structure
 *
 * Revision 1.2  1997/04/10 19:34:04  jhill
 * API changes
 *
 * Revision 1.1  1996/11/02 01:01:06  jhill
 * installed
 *
 *
 */


#ifndef casCoreClientIL_h
#define casCoreClientIL_h

#include "caServerIIL.h" // caServerI in line func 
#include "casCtxIL.h" // casEventSys in line func 

//
// casCoreClient::getCAS()
//
inline caServerI &casCoreClient::getCAS() const
{
	return *this->ctx.getServer();
}

//
// casCoreClient::installAsyncIO()
//
inline void casCoreClient::installAsyncIO(casAsyncIOI &ioIn)
{
	this->lock();
	this->ioInProgList.add(ioIn);
	this->unlock();
}

//
// casCoreClient::removeAsyncIO()
//
inline void casCoreClient::removeAsyncIO(casAsyncIOI &ioIn)
{
	this->lock();
	this->ioInProgList.remove(ioIn);
	this->ctx.getServer()->ioBlockedList::signal();
	this->unlock();
}

#endif // casCoreClientIL_h

